from __future__ import annotations

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    field_validator,
    field_serializer,
    ValidationInfo,
)
from nxtomomill.models.utils import convert_str_to_tuple
from nxtomomill.settings import Tomo
from ._context import extract_contexts, _Key, key_to_str


def _cast_keys_section_to_tuple(value):
    if isinstance(value, str):
        # convert "sa, id16a:sx" to ("sa", "id16a:sx")
        value: tuple = convert_str_to_tuple(value)

    def invalid_type():
        raise TypeError(
            f"Field '{value}' expects a string or an iterable of {_Key}, "
            f"got {type(value)!r}."
        )

    if not isinstance(value, tuple):
        invalid_type()

    res = []
    for key in value:
        if isinstance(key, str):
            key = extract_contexts(key)
        if not isinstance(key, _Key):
            invalid_type()
        res.append(key)
    return tuple(res)


class KeysSection(BaseModel):

    model_config = ConfigDict(validate_assignment=True, validate_by_name=True)

    valid_camera_names: tuple[_Key, ...] = Field(
        default=_cast_keys_section_to_tuple(Tomo.H5.VALID_CAMERA_NAMES),
        description="Valid camera names.",
    )
    rotation_angle_keys: tuple[_Key, ...] = Field(
        default=_cast_keys_section_to_tuple(Tomo.H5.ROT_ANGLE_KEYS),
        description="Keys for rotation angle.",
    )
    sample_x_keys: tuple[_Key, ...] = Field(
        default=_cast_keys_section_to_tuple(Tomo.H5.SAMPLE_X_KEYS),
        description="Keys for sample x translation.",
    )
    sample_y_keys: tuple[_Key, ...] = Field(
        default=_cast_keys_section_to_tuple(Tomo.H5.SAMPLE_Y_KEYS),
        description="Keys for sample y translation.",
    )
    translation_z_keys: tuple[_Key, ...] = Field(
        default=_cast_keys_section_to_tuple(Tomo.H5.TRANSLATION_Z_KEYS),
        description="Keys for translation in z.",
    )
    translation_y_keys: tuple[_Key, ...] = Field(
        default=_cast_keys_section_to_tuple(Tomo.H5.TRANSLATION_Y_KEYS),
        description="Keys for lateral translation. Also used to estimated center of rotation for half acquisition.",
    )
    translation_x_keys: tuple[_Key, ...] = Field(
        default=_cast_keys_section_to_tuple(Tomo.H5.TRANSLATION_X_KEYS),
        description="Keys for estimated center of rotation for half acquisition.",
    )

    diode_keys: tuple[_Key, ...] = Field(
        default=_cast_keys_section_to_tuple(Tomo.H5.DIODE_KEYS),
        description="Keys for diode.",
    )
    exposure_time_keys: tuple[_Key, ...] = Field(
        default=_cast_keys_section_to_tuple(Tomo.H5.ACQ_EXPO_TIME_KEYS),
        description="Keys for exposure time.",
    )
    sample_x_pixel_size_keys: tuple[_Key, ...] = Field(
        default=_cast_keys_section_to_tuple(Tomo.H5.SAMPLE_X_PIXEL_SIZE_KEYS),
        description="Keys for sample x pixel size.",
    )
    sample_y_pixel_size_keys: tuple[_Key, ...] = Field(
        default=_cast_keys_section_to_tuple(Tomo.H5.SAMPLE_Y_PIXEL_SIZE_KEYS),
        description="Keys for sample y pixel size.",
    )
    detector_x_pixel_size_keys: tuple[_Key, ...] = Field(
        default=_cast_keys_section_to_tuple(Tomo.H5.DETECTOR_X_PIXEL_SIZE_KEYS),
        description="Keys for detector x pixel size.",
    )
    detector_y_pixel_size_keys: tuple[_Key, ...] = Field(
        default=_cast_keys_section_to_tuple(Tomo.H5.DETECTOR_Y_PIXEL_SIZE_KEYS),
        description="Keys for detector y pixel size.",
    )
    sample_detector_distance_keys: tuple[_Key, ...] = Field(
        default=_cast_keys_section_to_tuple(Tomo.H5.SAMPLE_DETECTOR_DISTANCE_KEYS),
        description="Keys for sample to detector distance.",
        alias="sample_detector_distance",
    )
    source_sample_distance_keys: tuple[_Key, ...] = Field(
        default=_cast_keys_section_to_tuple(Tomo.H5.SOURCE_SAMPLE_DISTANCE_KEYS),
        description="Keys for source to sample distance.",
        alias="source_sample_distance",
    )
    machine_current_keys: tuple[_Key, ...] = Field(
        default=_cast_keys_section_to_tuple(Tomo.H5.MACHINE_CURRENT_KEYS),
        description="Keys for machine current.",
    )

    @field_validator(
        "valid_camera_names",
        "rotation_angle_keys",
        "sample_x_keys",
        "sample_y_keys",
        "translation_z_keys",
        "translation_y_keys",
        "translation_x_keys",
        "diode_keys",
        "exposure_time_keys",
        "sample_x_pixel_size_keys",
        "sample_y_pixel_size_keys",
        "detector_x_pixel_size_keys",
        "detector_y_pixel_size_keys",
        "sample_detector_distance_keys",
        "source_sample_distance_keys",
        "machine_current_keys",
        mode="wrap",
    )
    @classmethod
    def cast_keys_section_to_tuple(
        cls, value: str | tuple[_Key, ...], info: ValidationInfo
    ) -> tuple[_Key, ...]:
        return _cast_keys_section_to_tuple(value)

    @field_serializer(
        "valid_camera_names",
        "rotation_angle_keys",
        "sample_x_keys",
        "sample_y_keys",
        "translation_z_keys",
        "translation_y_keys",
        "translation_x_keys",
        "diode_keys",
        "exposure_time_keys",
        "sample_x_pixel_size_keys",
        "sample_y_pixel_size_keys",
        "detector_x_pixel_size_keys",
        "detector_y_pixel_size_keys",
        "sample_detector_distance_keys",
        "source_sample_distance_keys",
        "machine_current_keys",
        mode="wrap",
    )
    @classmethod
    def keys_with_contexts_to_tuple(
        cls, keys: tuple[_Key, ...], info: ValidationInfo
    ) -> tuple[str, ...]:
        if not keys:
            return tuple()
        elif not isinstance(keys, tuple):
            raise TypeError(f"keys is expected to be a tuple of {_Key}")
        # convert "sa, id16a:sx" to ("sa", "id16a:sx")
        return tuple([key_to_str(key) for key in keys])



"""Define models for existing (h52nx) context and helper functions."""

from __future__ import annotations

import re
from enum import Enum
from typing import Union
from pydantic import BaseModel, field_validator

"""Context can be given with settings keys to define a search context like
- 'bliss_tomo:sample_u' -> will use the dataset defined in 'bliss_tomo' to access sample_u information
- 'id16a:bliss_tomo:sample_u' -> If the dataset comes from id16a beamline will use the dataset defined in 'bliss_tomo' to access sample_u information
"""


class Proxy(Enum):
    """Determines if to access a dataset we read it directly or if we need to pass by a proxy like 'tomo_config'"""

    TOMO_CONFIG = "tomo_config"

    @classmethod
    def from_raw(cls, value: str):
        value = value.lower()
        if value == "tomoconfig":
            return Proxy.TOMO_CONFIG
        return cls(value)


class Beamline(Enum):
    id03 = "id03"
    bm05 = "bm05"
    bm18 = "bm18"
    id11 = "id11"
    id15 = "id15a"
    id16a = "id16a"
    id16b = "id16b"
    id17 = "id17"
    id19 = "id19"

    @classmethod
    def from_raw(cls, value: str):
        value = value.lower()
        value = value.replace("-", "")
        value = value.replace(":", "")
        value = value.replace("esrf", "")
        return cls(value)


_CONTEXT = Union[Proxy, Beamline]
_CONTEXT_CLASSES = {Proxy, Beamline}


class _Key(BaseModel):
    dataset: str
    contexts: set[_CONTEXT]

    @field_validator(
        "dataset",
        mode="plain",
    )
    @classmethod
    def check_dataset(cls, value: str) -> str:
        if ":" in value:
            raise ValueError(
                f"dataset name ({value}) is not expected to contain ':' char. Used as separator"
            )
        return value

    @staticmethod
    def resolve_dataset_operations(dataset_name: str) -> tuple[str, ...]:
        """Splits the dataset name to a chain of ('raw_dataset', operation,) ..."""
        # Use regex to split on operators (+, -, *, /) while keeping the operators
        tokens = re.split(r"([+\-*/])", dataset_name)
        # Remove empty strings and strip whitespace
        tokens = [token.strip() for token in tokens if token.strip()]
        return tokens


def extract_contexts(key: str, separator=":") -> _Key:
    """
    Split a given string into a dataset name and a set of context (beamline number, 'bliss_tomo'...)
    Order of the context doesn't matter but the dataset name must be at the end.
    """
    parts = key.split(separator)
    dataset_name = parts[-1]
    if len(parts) > 1:

        contexts = {_str_to_context(part) for part in parts[:-1]}
    else:
        contexts = set()
    if not dataset_name:
        raise ValueError(f"No dataset name provided in {key}")
    return _Key(dataset=dataset_name, contexts=contexts)


def _str_to_context(ctx: str) -> _CONTEXT:
    """convert a given string like 'ESRF-id11', 'tomo_config' to a context. Else raise an error"""
    for context in _CONTEXT_CLASSES:
        try:
            return context.from_raw(ctx)
        except ValueError:
            pass

    else:
        raise ValueError(f"Unrecognized context: {ctx}")


def key_to_str(key: _Key) -> str:
    """serialize instance of '_Key' to a string"""
    if not isinstance(key, _Key):
        raise TypeError(f"key is expected to be an instance of {_Key}")
    parts = [
        ctx.value
        for ctx in sorted(key.contexts, key=lambda context: isinstance(context, Proxy))
    ] + [
        key.dataset,
    ]
    return ":".join(parts)



from __future__ import annotations

import pint
import pytest
import h5py
import numpy

from nxtomomill.converter.hdf5.utils import (
    get_default_output_file,
    PROCESSED_DATA_DIR_NAME,
    RAW_DATA_DIR_NAME,
)
from nxtomomill.models.h52nx.h52nx import H52nxModel
from nxtomomill.converter.hdf5.acquisition.utils._context import (
    Proxy,
    _BeamlineContext,
    update_config_keys_from_beamline_context,
    update_config_keys_from_tomo_config_context,
    read_key,
    MissingValue,
)
from nxtomomill.converter.hdf5.acquisition.blisstomoconfig import TomoConfig
from nxtomomill.converter.hdf5.acquisition.utils._keys import _Key

_ureg = pint.UnitRegistry()


def test_get_default_output_file():
    """
    test the get_default_output_file function
    """
    # 1. simple test
    assert get_default_output_file("/my_tmp/path/file.h5") == "/my_tmp/path/file.nx"
    assert (
        get_default_output_file(f"/my_tmp/{RAW_DATA_DIR_NAME}/file.h5")
        == f"/my_tmp/{PROCESSED_DATA_DIR_NAME}/file.nx"
    )

    assert (
        get_default_output_file(f"/my_tmp/path/{RAW_DATA_DIR_NAME}/toto/file.h5")
        == f"/my_tmp/path/{PROCESSED_DATA_DIR_NAME}/toto/file.nx"
    )
    # note: _RAW_DATA_DIR_NAME part of the path but not a folder
    assert (
        get_default_output_file(f"/my_tmp/path_{RAW_DATA_DIR_NAME}/toto/file.h5")
        == f"/my_tmp/path_{RAW_DATA_DIR_NAME}/toto/file.nx"
    )

    # 2. advance test
    # 2.1 use case: '_RAW_DATA_DIR_NAME' is present twice in the path -> replace the deeper one
    assert (
        get_default_output_file(
            f"/my_tmp/{RAW_DATA_DIR_NAME}/path/{RAW_DATA_DIR_NAME}/toto/file.h5"
        )
        == f"/my_tmp/{RAW_DATA_DIR_NAME}/path/{PROCESSED_DATA_DIR_NAME}/toto/file.nx"
    )

    # 2.2 use case: contains both '_RAW_DATA_DIR_NAME' and '_PROCESSED_DATA_DIR_NAME' in the path
    assert (
        get_default_output_file(
            f"/my_tmp/{RAW_DATA_DIR_NAME}/path/{PROCESSED_DATA_DIR_NAME}/toto/file.h5"
        )
        == f"/my_tmp/{RAW_DATA_DIR_NAME}/path/{PROCESSED_DATA_DIR_NAME}/toto/file.nx"
    )

    assert (
        get_default_output_file(
            f"/my_tmp/{PROCESSED_DATA_DIR_NAME}/path/{RAW_DATA_DIR_NAME}/toto/file.h5"
        )
        == f"/my_tmp/{PROCESSED_DATA_DIR_NAME}/path/{PROCESSED_DATA_DIR_NAME}/toto/file.nx"
    )

    # 2.3 use case: expected output file is the input file. Make sure append '_nxtomo'
    assert (
        get_default_output_file("/my_tmp/path/file.nx") == "/my_tmp/path/file_nxtomo.nx"
    )


def test_update_config_keys_from_tomo_config_context():
    tomo_config = TomoConfig()
    # tomoconfig register motors as (real_motor_name, bliss_alias)
    tomo_config.sample_x = ("alias_sx", "sx")
    tomo_config.sample_y = ("alias_sy", "sy")
    tomo_config.translation_y = ("alias_ty", "ty")

    model = H52nxModel()
    model.sample_x_keys = (_Key(dataset="sample_x", contexts={Proxy.TOMO_CONFIG}),)
    model.sample_y_keys = (
        _Key(dataset="sample_y+translation_y", contexts={Proxy.TOMO_CONFIG}),
    )

    update_config_keys_from_tomo_config_context(model, tomo_config=tomo_config)
    assert model.sample_x_keys == (_Key(dataset="sx", contexts=set()),)
    assert model.sample_y_keys == (_Key(dataset="sy+ty", contexts=set()),)


def test_update_config_keys_from_beamline_context():
    model = H52nxModel()
    model.sample_x_keys = ("esrf-id11:sx", "ud", "id03:tx")

    update_config_keys_from_beamline_context(model, _BeamlineContext.id11)
    assert model.sample_x_keys == (
        _Key(dataset="sx", contexts={_BeamlineContext.id11}),
        _Key(dataset="ud", contexts=set()),
    )

    update_config_keys_from_beamline_context(model, None)
    # if no beamline found we want to get rid of all the keys with a beamline context
    assert model.sample_x_keys == (_Key(dataset="ud", contexts=set()),)


@pytest.fixture
def read_key_datasets(tmp_path):
    """
    Create an HDF5 file with datasets of coherent and incoherent sizes,
    some with units and some without.
    """
    # Create a temporary HDF5 file
    file_path = tmp_path / "test_datasets.h5"

    with h5py.File(file_path, "w") as f:
        # Create datasets with coherent sizes (same shape)
        f.create_dataset("coherent/dataset1", data=numpy.random.rand(10))
        f.create_dataset("coherent/dataset2", data=numpy.random.rand(10))

        # Create datasets with units (using pint)
        f.create_dataset("with_incoherent_units/dataset5", data=numpy.random.rand(10))
        f["with_incoherent_units/dataset5"].attrs["unit"] = "meter"

        f.create_dataset("with_incoherent_units/dataset6", data=numpy.random.rand(10))
        f["with_incoherent_units/dataset6"].attrs["unit"] = "second"

        # Create datasets without units
        f.create_dataset("without_units/dataset7", data=numpy.ones(10))
        f.create_dataset("without_units/dataset8", data=numpy.ones(10))

        # Create datasets with **DIS**
        f.create_dataset("missing_data/dataset9", data="*DIS*")
        f.create_dataset("missing_data/dataset10", data=numpy.random.random(10))

        # Create datasets with coherent units (but different)
        f.create_dataset("with_coherent_units/dataset11", data=numpy.ones(10))
        f["with_coherent_units/dataset11"].attrs["unit"] = "meter"

        f.create_dataset("with_coherent_units/dataset12", data=(numpy.ones(10) * 2))
        f["with_coherent_units/dataset12"].attrs["unit"] = "cm"
        f.create_dataset("with_coherent_units/dataset13", data=numpy.ones(10))
        f["with_coherent_units/dataset13"].attrs["unit"] = "cm"

        # Create datasets with coherent units (arrays with different size)
        f.create_dataset("with_different_nb_points/dataset14", data=numpy.ones(10))
        f["with_different_nb_points/dataset14"].attrs["unit"] = "meter"

        f.create_dataset("with_different_nb_points/dataset15", data=(numpy.ones(6) * 2))
        f["with_different_nb_points/dataset15"].attrs["unit"] = "cm"

        # Create datasets with coherent units (arrays and scalars)
        f.create_dataset("with_scalar_and_arrays/dataset16", data=numpy.ones(10))
        f["with_scalar_and_arrays/dataset16"].attrs["unit"] = "meter"

        f.create_dataset("with_scalar_and_arrays/dataset17", data=1.4)
        f["with_scalar_and_arrays/dataset17"].attrs["unit"] = "cm"

        f.create_dataset("with_scalar_and_arrays/dataset18", data=(numpy.ones(10) * 2))
        f["with_scalar_and_arrays/dataset18"].attrs["unit"] = "meter"

        # Create datasets with coherent units (arrays and scalars)
        f.create_dataset(
            "with_scalar_and_incoherent_arrays/dataset19", data=numpy.ones(10)
        )
        f["with_scalar_and_incoherent_arrays/dataset19"].attrs["unit"] = "meter"

        f.create_dataset("with_scalar_and_incoherent_arrays/dataset20", data=1.4)
        f["with_scalar_and_incoherent_arrays/dataset20"].attrs["unit"] = "cm"

        f.create_dataset(
            "with_scalar_and_incoherent_arrays/dataset21", data=numpy.ones(9)
        )
        f["with_scalar_and_incoherent_arrays/dataset21"].attrs["unit"] = "meter"

    return file_path


def test_read_key(read_key_datasets, caplog):
    with h5py.File(read_key_datasets) as h5f:
        # test with dataset not existing
        with pytest.raises(KeyError):
            read_key(
                key="dataset3+dataset4",
                node=h5f["coherent"],
                info_retrieve="",
                expected_unit=_ureg.meter,
            )

        # test with dataset not dump
        with pytest.raises(MissingValue):
            read_key(
                key="dataset9+dataset10",
                node=h5f["missing_data"],
                info_retrieve="",
                expected_unit=_ureg.meter,
            )

        # test with dataset with coherent size and no unit
        quantity: pint.Quantity = read_key(
            key="dataset7+dataset8",
            node=h5f["without_units"],
            info_retrieve="",
            expected_unit=_ureg.centimeter,
        )
        numpy.testing.assert_array_equal(
            quantity.to_base_units(), numpy.array([0.02] * 10)
        )

        # test with dataset with coherent size but incoherent units
        with pytest.raises(pint.errors.DimensionalityError):
            quantity: pint.Quantity = read_key(
                key="dataset5+dataset6",
                node=h5f["with_incoherent_units"],
                info_retrieve="",
                expected_unit=_ureg.centimeter,
            )

        # test with dataset with coherent size and coherent units
        quantity: pint.Quantity = read_key(
            key="dataset11+dataset12-dataset13",
            node=h5f["with_coherent_units"],
            info_retrieve="",
            expected_unit=_ureg.centimeter,
        )
        numpy.testing.assert_array_equal(
            quantity.to_base_units(), numpy.array([1.01] * 10)
        )

        # test arrays with incoherent size
        key = "dataset14+dataset15"
        expected_warning = (
            "nxtomomill.converter.hdf5.acquisition.utils._context",
            30,
            f"Incoherent number of points founds for info when accessing '{key}'.",
        )
        assert expected_warning not in caplog.record_tuples
        quantity: pint.Quantity = read_key(
            key=key,
            node=h5f["with_different_nb_points"],
            info_retrieve="info",
            expected_unit=_ureg.centimeter,
        )
        numpy.testing.assert_array_equal(
            quantity.to_base_units(), numpy.array([1.02] * 6)
        )
        assert expected_warning in caplog.record_tuples

        # test arrays with scalar
        quantity: pint.Quantity = read_key(
            key="dataset16+dataset17-dataset18",
            node=h5f["with_scalar_and_arrays"],
            info_retrieve="",
            expected_unit=_ureg.centimeter,
        )
        numpy.testing.assert_array_equal(
            quantity.to_base_units(), numpy.array([1 + 0.014 - 2] * 10)
        )
        # test incoherent arrays with scalar
        quantity: pint.Quantity = read_key(
            key="dataset19-dataset20+dataset21",
            node=h5f["with_scalar_and_incoherent_arrays"],
            info_retrieve="",
            expected_unit=_ureg.centimeter,
        )
        numpy.testing.assert_array_equal(
            quantity.to_base_units(), numpy.array([1 - 0.014 + 1] * 9)
        )



from __future__ import annotations

import os
import pint
import numpy
import h5py

from nxtomo.application.nxtomo import NXtomo
from nxtomo.nxobject.nxdetector import ImageKey

from nxtomomill.tests.datasets import GitlabDataset
from nxtomomill import converter
from nxtomomill.models.h52nx import H52nxModel
from nxtomomill.models.h52nx.keys_section import _cast_keys_section_to_tuple

_ureg = pint.get_application_registry()


def test_holotomo_conversion_id16a(tmp_path):
    """
    test conversion on one id16a use case

    check x, y and z positions
    warning: bliss-tomo is using esrf coordinate system when
    NXtomo is using the McStas one. see https://tomotools.gitlab-pages.esrf.fr/tomwer/development/referential.html for more details.
    on esrf referential we expect the following (from peter):

        translation_x = sx (to keep things simple)
        translation_y = sy + spy
        translation_z = sz + spz
        sample_x = sax
        sample_y = say

    .. warning:: In 'tomoconfig'

    'translation_y is defined as 'sy' instead of sy + spy
    'translation_z is defined as 'sz' instead of sz + spz

    So this is why we need to overwrite the settings
    """
    dataset_file = os.path.join(
        "h5_datasets", "holotomo", "id16a", "SiemensStar_HT_025nm_0002.h5"
    )
    input_file = GitlabDataset.get_dataset(dataset_file)
    assert os.path.exists(input_file)
    output_file = os.path.join(tmp_path, "nexus_file.nx")

    assert not os.path.exists(output_file), "output_file exists already"
    configuration = H52nxModel()

    configuration.output_file = output_file
    configuration.input_file = input_file
    configuration.raises_error = False
    configuration.no_master_file = False
    configuration.translation_y_keys = _cast_keys_section_to_tuple(("sy+spy",))
    configuration.translation_z_keys = _cast_keys_section_to_tuple(("sz+spz",))

    result = converter.from_h5_to_nx(configuration=configuration)
    assert os.path.exists(output_file)

    # check we have our 2 NXtomo are converted with (two distances).
    assert len(result) == 2
    # expected:
    nx_tomos = [NXtomo().load(file_path, entry) for (file_path, entry) in result]
    n_proj_per_nx_tomo = 1000
    n_frames = n_proj_per_nx_tomo + 20 * 3 + 3
    nx_tomo = nx_tomos[0]
    # check instrument
    assert len(nx_tomo.instrument.detector.image_key_control) == n_frames
    # image key
    numpy.testing.assert_array_equal(
        nx_tomo.instrument.detector.image_key_control,
        numpy.concatenate(
            (
                [ImageKey.DARK_FIELD] * 20,
                [ImageKey.FLAT_FIELD] * 20,
                [ImageKey.PROJECTION] * n_proj_per_nx_tomo,
                [ImageKey.ALIGNMENT] * 3,
                [ImageKey.FLAT_FIELD] * 20,
            )
        ),
    )

    # sequence number
    numpy.testing.assert_array_equal(
        nx_tomo.instrument.detector.sequence_number, numpy.arange(0, n_frames)
    )

    # detector pixel size
    numpy.testing.assert_almost_equal(
        nx_tomo.instrument.detector.x_pixel_size.magnitude,
        (2.95203 * _ureg.micrometer).magnitude,
        decimal=5,
    )
    numpy.testing.assert_almost_equal(
        nx_tomo.instrument.detector.y_pixel_size.magnitude,
        (2.95203 * _ureg.micrometer).magnitude,
        decimal=5,
    )

    assert numpy.isclose(
        nx_tomo.instrument.detector.distance, 1206.69 * _ureg.millimeter
    )
    # check energy
    assert nx_tomo.energy == 17.1 * _ureg.keV
    # check sample
    numpy.testing.assert_almost_equal(
        nx_tomo.sample.x_pixel_size.magnitude,
        (0.025 * _ureg.micrometer).magnitude,
        decimal=5,
    )
    numpy.testing.assert_almost_equal(
        nx_tomo.sample.y_pixel_size.magnitude,
        (0.025 * _ureg.micrometer).magnitude,
        decimal=5,
    )

    # check source
    numpy.testing.assert_almost_equal(
        nx_tomo.instrument.source.distance.magnitude,
        (-10.3065 * _ureg.millimeter).magnitude,
        decimal=4,
    )

    with h5py.File(input_file, mode="r") as h5f:
        scan_prj_index = 5
        sx = h5f[f"{scan_prj_index}.1/instrument/positioners/sx"][()] * _ureg.millimeter
        sy = h5f[f"{scan_prj_index}.1/instrument/positioners/sy"][()] * _ureg.millimeter
        spy = (
            h5f[f"{scan_prj_index}.1/instrument/positioners/spy"][()] * _ureg.micrometer
        )
        sz = h5f[f"{scan_prj_index}.1/instrument/positioners/sz"][()] * _ureg.millimeter
        spz = (
            h5f[f"{scan_prj_index}.1/instrument/positioners/spz"][()] * _ureg.micrometer
        )
        sax = (
            h5f[f"{scan_prj_index}.1/instrument/positioners/sax"][()] * _ureg.millimeter
        )
        say = (
            h5f[f"{scan_prj_index}.1/instrument/positioners/say"][()] * _ureg.millimeter
        )

    proj_index = numpy.arange(40, n_proj_per_nx_tomo + 40)

    numpy.testing.assert_almost_equal(
        nx_tomo.sample.x_translation.to_base_units().magnitude[proj_index],
        (sy + spy + say).to_base_units().magnitude,
    )
    numpy.testing.assert_almost_equal(
        nx_tomo.sample.y_translation.to_base_units().magnitude[proj_index],
        (sz + spz).to_base_units().magnitude,
    )
    numpy.testing.assert_almost_equal(
        (nx_tomo.sample.z_translation).to_base_units().magnitude[proj_index],
        (sx + sax).to_base_units().magnitude,
    )


def test_holotomo_conversion_id16b(tmp_path):
    """
    Test conversion of files from
    """
    dataset_file = os.path.join(
        "h5_datasets", "holotomo", "id16b", "20250711", "B3P7_sponge_pco2_ht_25nm.h5"
    )
    input_file = GitlabDataset.get_dataset(dataset_file)
    assert os.path.exists(input_file)
    output_file = os.path.join(tmp_path, "nexus_file.nx")

    configuration = H52nxModel()

    configuration.output_file = output_file
    configuration.input_file = input_file
    configuration.raises_error = False
    configuration.no_master_file = False

    result = converter.from_h5_to_nx(configuration=configuration)
    assert os.path.exists(output_file)

    # check we have our 4 NXtomo are converted with (two distances).
    assert len(result) == 4
    # expected:
    nx_tomos = [NXtomo().load(file_path, entry) for (file_path, entry) in result]
    n_proj_per_nx_tomo = 3203
    n_frames = n_proj_per_nx_tomo + (50 + 51) * 2 + 5

    nx_tomo = nx_tomos[0]
    # check instrument
    assert len(nx_tomo.instrument.detector.image_key_control) == n_frames
    # image key
    numpy.testing.assert_array_equal(
        nx_tomo.instrument.detector.image_key_control,
        numpy.concatenate(
            (
                [ImageKey.DARK_FIELD] * 50,
                [ImageKey.FLAT_FIELD] * 51,
                [ImageKey.PROJECTION] * n_proj_per_nx_tomo,
                [ImageKey.ALIGNMENT] * 5,
                [ImageKey.FLAT_FIELD] * 51,
                [ImageKey.DARK_FIELD] * 50,
            )
        ),
    )

    # sequence number
    numpy.testing.assert_array_equal(
        nx_tomo.instrument.detector.sequence_number, numpy.arange(0, n_frames)
    )

    # detector pixel size
    numpy.testing.assert_almost_equal(
        nx_tomo.instrument.detector.x_pixel_size.magnitude,
        (0.65 * _ureg.micrometer).magnitude,
        decimal=5,
    )
    numpy.testing.assert_almost_equal(
        nx_tomo.instrument.detector.y_pixel_size.magnitude,
        (0.65 * _ureg.micrometer).magnitude,
        decimal=5,
    )

    assert numpy.isclose(
        nx_tomo.instrument.detector.distance, 723.908 * _ureg.millimeter
    )
    # check energy
    assert nx_tomo.energy == 29.6 * _ureg.keV
    # check sample
    numpy.testing.assert_almost_equal(
        nx_tomo.sample.x_pixel_size.magnitude,
        (0.0250011 * _ureg.micrometer).magnitude,
        decimal=5,
    )
    numpy.testing.assert_almost_equal(
        nx_tomo.sample.y_pixel_size.magnitude,
        (0.0250011 * _ureg.micrometer).magnitude,
        decimal=5,
    )

    # check source
    numpy.testing.assert_almost_equal(
        nx_tomo.instrument.source.distance.magnitude,
        (-28.9577 * _ureg.millimeter).magnitude,
        decimal=4,
    )

    with h5py.File(input_file, mode="r") as h5f:
        scan_prj_index = 4
        sx = h5f[f"{scan_prj_index}.1/instrument/positioners/sx"][()] * _ureg.millimeter
        sy = h5f[f"{scan_prj_index}.1/instrument/positioners/sy"][()] * _ureg.millimeter
        sz = h5f[f"{scan_prj_index}.1/instrument/positioners/sz"][()] * _ureg.millimeter

    proj_index = numpy.arange(101, n_proj_per_nx_tomo + 101)

    numpy.testing.assert_almost_equal(
        nx_tomo.sample.x_translation.to_base_units().magnitude[proj_index],
        sy.to_base_units().magnitude,
    )
    numpy.testing.assert_almost_equal(
        nx_tomo.sample.y_translation.to_base_units().magnitude[proj_index],
        sz.to_base_units().magnitude,
    )
    numpy.testing.assert_almost_equal(
        (nx_tomo.sample.z_translation).to_base_units().magnitude[proj_index],
        sx.to_base_units().magnitude,
    )



"""Define existing (h52nx) contexts and functions to handle them."""

from __future__ import annotations

import logging
import h5py
import numpy
import pint
from typing import TYPE_CHECKING, Callable
from silx.io.utils import h5py_read_dataset

from nxtomomill.models.h52nx import H52nxModel
from nxtomomill.models.h52nx._context import _Key, Proxy, Beamline as _BeamlineContext
from nxtomomill.utils.hdf5 import get_dataset_unit

_logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from nxtomomill.converter.hdf5.acquisition.blisstomoconfig import TomoConfig


_MODEL_FIELDS_WITH_KEYS = tuple(
    {
        key: value
        for key, value in H52nxModel.model_fields.items()
        if value.annotation == tuple[_Key, ...]
    }.keys()
)
"""Keys that can be contain a context == field that allows Field as type"""


def update_config_keys_from_tomo_config_context(
    configuration: H52nxModel, tomo_config: TomoConfig
):
    """Replace the tomo_config:xxx config by the dataset name contained in 'tomo_config'"""

    def map_dataset(dataset_name: str) -> str:
        """map a dataset from tomoconfig to it's raw name.

        ..warning:: tomoconfig dataset can store several dataset (alias, raw motor name). We are interested on the last only.
        """
        datasets = getattr(tomo_config, dataset_name)
        if datasets is None:
            raise KeyError(f"Cannot found tomo_config key: `{dataset_name}`")
        assert isinstance(
            datasets, (tuple, list)
        ), f"dataset is expected to be a tuple. Got {type(datasets)}"
        return datasets[-1]

    def apply_tomo_config_mapping(key: _Key) -> tuple[_Key, ...]:
        """
        Apply proxy / mapping from based on values contained in tomo_config.
        Example: 'sample_y+translation_y' becomes 'sy+ty' if tomo_config.sample_y == sy and tomo_config.translation_y == ty
        """
        split_dataset_and_operations = _Key.resolve_dataset_operations(key.dataset)
        mapping = [
            map_dataset(split_dataset_and_operations[0]),
        ]
        if len(split_dataset_and_operations) > 1:
            operators = split_dataset_and_operations[1::2]
            datasets = split_dataset_and_operations[2::2]
            for operation, dataset in zip(operators, datasets):
                mapping.extend((operation, map_dataset(dataset_name=dataset)))
        return "".join(mapping)

    def update_key(key: _Key) -> tuple[_Key, ...]:
        """One key return the original keys or two keys from tomoconfig: alias and motor"""
        if Proxy.TOMO_CONFIG in key.contexts:
            try:
                real_dataset_name = apply_tomo_config_mapping(key=key)
            except KeyError as e:
                _logger.error(e)
                # if the mapping cannot done with tomo_config log an error and remove the key from the list: consider it treated and invalid.
                return ()
            else:
                return (
                    _Key(
                        dataset=real_dataset_name,
                        contexts=key.contexts - {Proxy.TOMO_CONFIG},
                    ),
                )
        return (key,)

    update_configuration_keys(configuration, update_key)


def remove_config_keys_with_tomo_config_context(configuration: H52nxModel):
    """Remove all keys with a 'tomoconfig' context"""

    def update_key(key: _Key) -> tuple[_Key | None, ...]:
        if Proxy.TOMO_CONFIG in key.contexts:
            return (None,)
        else:
            return (key,)

    update_configuration_keys(configuration, update_key)


def update_config_keys_from_beamline_context(
    configuration: H52nxModel, beamline: _BeamlineContext | None
):
    """Function to filter settings according to the beamline"""

    def update_key(key: _Key) -> tuple[_Key | None, ...]:
        if not has_beamline_context(key) or (
            has_beamline_context(key) and beamline in key.contexts
        ):
            return (key,)
        else:
            return (None,)

    update_configuration_keys(configuration, update_key)


def has_beamline_context(key: _Key) -> bool:
    """Check if some beamline context are defined within the key"""
    for context in key.contexts:
        if context in _BeamlineContext:
            return True
    return False


def update_configuration_keys(
    configuration: H52nxModel,
    key_transform: Callable[[_Key]],
    fields: tuple[str, ...] = _MODEL_FIELDS_WITH_KEYS,
) -> None:
    """
    Generalized function to update keys in configuration fields.
    """

    def update_keys(keys):
        new_keys = []
        for key in keys:
            result = key_transform(key)
            new_keys.extend(result)
        return tuple(filter(None, new_keys))

    for keys_as_str in fields:
        keys_values = getattr(configuration, keys_as_str)
        setattr(configuration, keys_as_str, update_keys(keys_values))


class MissingValue(ValueError):
    """Raised if a dataset contains 'DIS' - not saved"""


def read_key(
    key: str,
    node: h5py.Group,
    expected_unit: pint.Unit,
    info_retrieve: str,
) -> pint.Quantity:
    """
    Util function to read a dataset from it's name.
    The dataset can be a composition of datasets and operations like 'sx+tx-ax'.
    In this case we first read the three dataset then apply sx+tx and finally tx-ax

    .. note::

        Unit of each dataset will be read and take into account. Dataset can link to scalar values or 1D arrays.
        If arrays have incoherent size (some points not recorded for example) then longest datasets will be cropped to match shortest dataset. End of the dataset will be ignored.
        A log message will be added.

    .. note::

        key can contain one of the following operator:

        * '+' for addition between two datasets
        * '-' for subtraction between two datasets

    .. note::

        If dataset contains '*DIS*'. Can happen for old dataset if the dataset has not been recorded. Will raise a MissingValue.

    """
    parts = _Key.resolve_dataset_operations(key)
    dataset_names = parts[::2]
    if len(parts) > 1:
        operators = parts[1::2]

    dataset_values = {}
    dataset_units = {}
    for dataset_name in dataset_names:
        values = h5py_read_dataset(node[dataset_name])
        dataset_values[dataset_name] = values
        dataset_units[dataset_name] = get_dataset_unit(
            dataset=node[dataset_name],
            default=expected_unit,
            from_dataset=f"{node[dataset_name].name} (reading {info_retrieve})",
        )
        # skip values containing '*DIS*'
        if isinstance(values, str) and values == "*DIS*":
            raise MissingValue(f"dataset {dataset_name} has not been recorded")

    first_dataset = dataset_names[0]
    result_unit = dataset_units[first_dataset] or expected_unit
    result = dataset_values[first_dataset] * result_unit

    # Map operation strings to NumPy functions
    OPERATION_MAP = {
        "+": numpy.add,
        "-": numpy.subtract,
    }

    if len(dataset_names) > 1:
        for operator, dataset_name in zip(operators, dataset_names[1:]):
            dataset_unit = dataset_units[dataset_name] or expected_unit
            dataset_quantity = dataset_values[dataset_name] * dataset_unit
            if (
                not numpy.isscalar(result.magnitude)
                and not numpy.isscalar(dataset_quantity.magnitude)
                and len(result.magnitude) != len(dataset_quantity.magnitude)
            ):
                # crop dataset if have an incoherent size.
                _logger.warning(
                    f"Incoherent number of points founds for {info_retrieve} when accessing '{key}'."
                )
                nb_pts = min(len(result), len(dataset_quantity))
                result = result[:nb_pts]
                dataset_quantity = dataset_quantity[:nb_pts]
            result = OPERATION_MAP[operator](result, dataset_quantity)

    return result



from __future__ import annotations

from nxtomomill.models.h52nx._context import (
    extract_contexts,
    Beamline,
    Proxy,
    key_to_str,
    _Key,
)
import pytest


def test_extract_contexts():
    assert extract_contexts("id16a:tomo_config:sample_u") == _Key(
        dataset="sample_u",
        contexts={Proxy.TOMO_CONFIG, Beamline.id16a},
    )
    assert extract_contexts("sample_u") == _Key(
        dataset="sample_u",
        contexts=set(),
    )
    assert extract_contexts("id19:sample_u") == _Key(
        dataset="sample_u",
        contexts={Beamline.id19},
    )

    with pytest.raises(ValueError):
        extract_contexts("id16a:")
    with pytest.raises(ValueError):
        extract_contexts("tomo_config:")
    with pytest.raises(ValueError):
        extract_contexts("unknown:sample_u")


def test_key_to_str():
    assert (
        key_to_str(
            _Key(
                dataset="toto",
                contexts={
                    Beamline.id03,
                    Proxy.TOMO_CONFIG,
                },
            )
        )
        == f"{Beamline.id03.value}:{Proxy.TOMO_CONFIG.value}:toto"
    )
    assert (
        key_to_str(
            _Key(
                dataset="toto",
                contexts={
                    Proxy.TOMO_CONFIG,
                },
            )
        )
        == f"{Proxy.TOMO_CONFIG.value}:toto"
    )
    assert (
        key_to_str(
            _Key(
                dataset="toto",
                contexts=set(),
            )
        )
        == "toto"
    )
    with pytest.raises(TypeError):
        key_to_str(
            {
                "dataset": "toto",
                "contexts": set(),
            }
        )



"""A set of utils related to keys management"""

from .....models.h52nx._context import _Key


def keys_to_str_tuple(keys: tuple[_Key]) -> tuple[str]:
    for key in keys:
        if not isinstance(key, _Key):
            raise TypeError(f"key should be an instance of {_Key}. Got {type(key)}")
    return tuple([key.dataset for key in keys])



"""pint utils"""

from __future__ import annotations

import pint
import logging
from pint.errors import UndefinedUnitError

_logger = logging.getLogger(__name__)

_ureg = pint.get_application_registry()


VALID_CURRENT_VALUES: tuple[str] = [
    str(unit) for unit in (_ureg.ampere, _ureg.kiloampere, _ureg.milliampere)
]
VALID_ENERGY_VALUES: tuple[str] = [
    str(unit)
    for unit in (_ureg.joule, _ureg.eV, _ureg.keV, _ureg.meV, _ureg.GeV, _ureg.kJ)
]
VALID_METRIC_VALUES: tuple[str] = [
    str(unit)
    for unit in (
        _ureg.nanometer,
        _ureg.micrometer,
        _ureg.millimeter,
        _ureg.centimeter,
        _ureg.meter,
    )
]


def get_unit(unit: str, default: pint.Unit, from_dataset: str) -> pint.Unit:
    """
    Convert given unit as an str to a pint unit. Else fallback to 'default'
    """
    if not isinstance(unit, str):
        raise TypeError(f"'unit' is expected to be a {str}. Got {type(unit)}")
    if not isinstance(default, pint.Unit):
        raise TypeError(
            f"'default' is expected to be a {pint.Unit}. Got {type(default)}"
        )

    # special cases (some string that can exists and that are not handled by pint)
    if unit.lower() in ("mev", "megaelectronvolt"):
        unit = _ureg.keV * 1e3
    elif unit.lower() in ("gev", "gigaelectronvolt"):
        unit = _ureg.keV * 1e6
    elif unit == "microns":
        unit = "um"

    try:
        unit = _ureg(unit)
    except UndefinedUnitError:
        _logger.warning(
            f"Undefined unit: {unit} from {from_dataset}. Fallback on default unit {default}"
        )
        return default
    else:
        return unit


def sum_optional_quantities(
    a1: pint.Quantity | None, a2: pint.Quantity | None
) -> pint.Quantity | None:
    """Sum two variables that are quantities or None. If one is None we silently ignore it."""
    assert isinstance(a1, (pint.Quantity, type(None)))
    assert isinstance(a2, (pint.Quantity, type(None)))

    if a1 is None and a2 is None:
        return None
    if a1 is not None and a2 is not None:
        return a1 + a2
    elif a1 is not None:
        return a1
    else:
        return a2



# coding: utf-8

import os

from tomoscan.esrf.mock import MockEDF

from nxtomomill.app.edf2nx import main


def test_edf2nx_application(tmp_path):
    """test nxtomomill edf2nx input_file output_file"""
    edf_acq_path = tmp_path / "acquisition"
    edf_acq_path.mkdir()
    edf_acq_path = str(edf_acq_path)
    n_proj = 10
    MockEDF(
        scan_path=edf_acq_path,
        n_radio=n_proj,
        n_ini_radio=n_proj,
    )
    output_file = os.path.join(edf_acq_path, "nexus_file.nx")

    assert not os.path.exists(output_file), "output_file exists already"
    main(["edf2nx", edf_acq_path, output_file])
    assert os.path.exists(output_file), "output_file doesn't exists"



from __future__ import annotations

from pydantic import Field, field_serializer, field_validator, ConfigDict

from nxtomo.nxobject.nxdetector import FieldOfView

from nxtomomill.models.utils import filter_str_def, convert_str_to_bool
from ..base.general_section import GeneralSection as _GeneralSectionBase


class GeneralSection(_GeneralSectionBase):
    model_config = ConfigDict(validate_assignment=True, validate_by_name=True)

    input_file: str | None = Field(default=None, description="Input file path.")
    raises_error: bool = Field(
        default=False, description="Raise an error when encountered."
    )
    no_input: bool = Field(
        default=True,
        description="If True, do not ask for user inputs if missing information.",
    )
    single_file: bool = Field(
        default=False, description="Create a single file for all sequences."
    )
    no_master_file: bool = Field(
        default=True, description="Avoid creating the master file."
    )
    field_of_view: FieldOfView | None = Field(
        default=None,
        description="Force output to be a 'Full' or 'Half' acquisition. Else determine from existing metadata",
    )
    create_control_data: bool = Field(
        default=True, description="Generate control/data - filled with machine current."
    )
    check_tomo_n: bool | None = Field(
        default=None,
        description="Check 'tomo_n' metadata once the conversion is done. By default will be done for all except scan build from 'scan_data'.",
    )
    rotation_is_clockwise: bool | None = Field(
        default=None,
        description="Force rotation clockwise or not. If not set will read the information from the bliss-tomo 'tomoconfig' group. Else consider rotation is counterclockwise.",
    )
    mechanical_lr_flip: bool = Field(
        default=False,
        description="Detector image is flipped **left-right** for mechanical reasons that are not propagated with bliss-tomo hdf5 metadata (mirror,..).",
    )
    mechanical_ud_flip: bool = Field(
        default=False,
        description="Detector image is flipped **up-down** for mechanical reasons that are not propagated with bliss-tomo hdf5 metadata (mirror,..).",
    )

    @property
    def request_input(self) -> bool:
        return not self.no_input

    @request_input.setter
    def request_input(self, request: bool):
        self.no_input = request

    @field_validator(
        "check_tomo_n",
        "rotation_is_clockwise",
        mode="before",
    )
    @classmethod
    def cast_general_section_to_bool_or_none(
        cls, value: str | None | bool
    ) -> bool | None:
        if isinstance(value, str):
            value = filter_str_def(value)
        if value in (None, "None", ""):
            return None
        elif value in ("True", True, "1"):
            return True
        elif value in ("False", False, "0"):
            return False
        else:
            raise ValueError

    @field_serializer(
        "check_tomo_n",
        "rotation_is_clockwise",
        when_used="always",
    )
    @classmethod
    def serialize_check_tomo_n(cls, value: None | bool) -> str:
        if value is None:
            return None
        else:
            return value

    @field_validator(
        "field_of_view",
        mode="plain",
    )
    @classmethod
    def cast_to_field_of_view(cls, value: str | FieldOfView) -> FieldOfView | None:
        if value in (None, ""):
            return None
        return FieldOfView(value)

    @field_serializer(
        "field_of_view",
        when_used="always",
    )
    @classmethod
    def serialize_field_of_view(cls, field_of_view: FieldOfView | None) -> str:
        if field_of_view is None:
            return None
        else:
            return field_of_view.value

    @field_validator(
        "input_file",
        mode="before",
    )
    @classmethod
    def cast_general_section_to_str(cls, value: str | None) -> str | None:
        return filter_str_def(value)

    @field_validator(
        "no_input",
        "single_file",
        "no_master_file",
        "create_control_data",
        "mechanical_lr_flip",
        "mechanical_ud_flip",
        mode="before",
    )
    @classmethod
    def cast_general_section_to_bool(cls, value: str | bool | None) -> bool:
        return convert_str_to_bool(value)
